@@ -100,7 +100,7 @@ class UpdateOrder(Enum):
 
 def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None, input=None, topology_rule=None,
            perturbation=None, past_conditions=None, update_order=UpdateOrder.ACTIVITIES_FIRST, copy_network=True,
-           compression=False, persist_activities=True, persist_network=True):
+           compression=False, persist_activities=True, persist_network=True, memoize=False, memoization_key=None):
     """
     Evolves the given network with the given initial conditions, for the specified number of timesteps, using the given
     activity and topology rules. Note that if A(t) is the network at timestep t, and S(t) is the network activity vector
@@ -149,6 +149,14 @@ def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None,
     :param compression: whether to compress the network when it is persisted in the State (default is False)
     :param persist_activities: whether to persist the activities in the State (default is True)
     :param persist_network: whether to persist the network in the State (default is True)
+    :param memoize: whether to use memoization during evolution; if True, then the result of applying the rule for a
+                    a given input will be cached, and used on subsequent invocations of the rule; this can result
+                    in a significant improvement to execution speed if the rule is expensive to invoke; NOTE: this
+                    should only be set to True for rules which do not store any state upon invocation (default is False)
+    :param memoization_key: a user-specified key for memoization; by default, an appropriate memoization key is
+                            used depending on the presence or absence of a rotation system on the network;
+                            must be a `MemoizationKey` instance; this argument is relevant only when `memoize` is `True`
+                            (default is None)
     :return: a trajectory, which is a list of States, where each State defines the network and its activities for a
              particular timestep
     """
@@ -171,6 +179,15 @@ def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None,
     prev_activities = initial_conditions
     prev_network = network
 
+    memo_table = {}
+    if memoize and memoization_key is None:
+        memoization_key = _UnorderedMemoizationKey() if network.rotation_system is None else _OrderedMemoizationKey()
+
+    # create a neighbourhood map for efficiency if the network never changes (i.e. no topology rule)
+    neighbourhood_map = None
+    if topology_rule is None:
+        neighbourhood_map = _get_neighbourhood_map(network)
+
     t = 1
     while True:
         inp = input_fn(t, prev_activities, prev_network)
@@ -183,25 +200,28 @@ def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None,
 
         if update_order is UpdateOrder.ACTIVITIES_FIRST:
             added_nodes, removed_nodes = evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
-                                                           trajectory, prev_network, past,
-                                                           perturbation, persist_activities)
+                                                           trajectory, prev_network, past, perturbation, persist_activities,
+                                                           memoize, memo_table, neighbourhood_map, memoization_key)
             curr_network = evolve_topology(topology_rule, t, curr_activities, prev_network,
                                            trajectory, copy_network, persist_network, added_nodes, removed_nodes)
+            if added_nodes or removed_nodes:
+                # re-build the neighbourhood map, since the network has changed
+                neighbourhood_map = _get_neighbourhood_map(curr_network)
 
         elif update_order is UpdateOrder.TOPOLOGY_FIRST:
             curr_network = evolve_topology(topology_rule, t, prev_activities, prev_network,
                                            trajectory, copy_network, persist_network)
             # added and removed nodes are ignore in this case
-            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
-                              trajectory, curr_network, past, perturbation, persist_activities)
+            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities, trajectory, curr_network, past,
+                              perturbation, persist_activities, memoize, memo_table, neighbourhood_map, memoization_key)
 
         elif update_order is UpdateOrder.SYNCHRONOUS:
             # TODO create test
             curr_network = evolve_topology(topology_rule, t, prev_activities, prev_network,
                                            trajectory, copy_network, persist_network)
             # added and removed nodes are ignore in this case
-            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
-                              trajectory, prev_network, past, perturbation, persist_activities)
+            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities, trajectory, prev_network, past,
+                              perturbation, persist_activities, memoize, memo_table, neighbourhood_map, memoization_key)
 
         else:
             raise Exception("unsupported update_order: %s" % update_order)
@@ -217,12 +237,14 @@ def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None,
 
 
 def evolve_activities(activity_rule, t, inp, curr_activities, prev_activities, trajectory, network,
-                      past, perturbation, persist_activities):
+                      past, perturbation, persist_activities, memoize, memoization_table, neighbourhood_map,
+                      memoization_key):
     added_nodes = []
     removed_nodes = []
     if activity_rule:
         added_nodes, removed_nodes = do_activity_rule(t, inp, curr_activities, prev_activities, network,
-                                                      activity_rule, past, perturbation)
+                                                      activity_rule, past, perturbation, memoize, memoization_table,
+                                                      neighbourhood_map, memoization_key)
 
         if added_nodes:
             for state, outgoing_links, node_label in added_nodes:
@@ -234,21 +256,25 @@ def evolve_activities(activity_rule, t, inp, curr_activities, prev_activities, t
     return added_nodes, removed_nodes
 
 
-def do_activity_rule(t, inp, curr_activities, prev_activities, network, activity_rule, past, perturbation):
+def do_activity_rule(t, inp, curr_activities, prev_activities, network, activity_rule, past, perturbation,
+                     memoize, memoization_table, neigh_map, memoization_key):
     added_nodes = []
     removed_nodes = []
 
     for node_label in network.nodes:
         incoming_connections = network.in_edges(node_label)
-        neighbour_labels = [i for i in incoming_connections]
+        neighbour_labels = _get_neighbour_labels(node_label, neigh_map, network, incoming_connections)
         current_activity = prev_activities[node_label]
         neighbourhood_activities = [prev_activities[neighbour_label] for neighbour_label in neighbour_labels]
         node_in = None if inp == "__timestep__" else inp[node_label] if _is_indexable(inp) else inp
         ctx = NodeContext(node_label, t, prev_activities, neighbour_labels, neighbourhood_activities,
                           incoming_connections, current_activity, past, node_in)
 
-        new_activity = activity_rule(ctx)
-        new_activity = check_np(new_activity)
+        if memoize:
+            new_activity = _get_memoized(activity_rule, ctx, memoization_table, memoization_key)
+        else:
+            new_activity = activity_rule(ctx)
+            new_activity = check_np(new_activity)
 
         if ctx.added_nodes:
             added_nodes.extend(ctx.added_nodes)
@@ -265,6 +291,69 @@ def do_activity_rule(t, inp, curr_activities, prev_activities, network, activity
     return added_nodes, removed_nodes
 
 
+def _get_neighbour_labels(node_label, neighbourhood_map, network, incoming_connections):
+    if neighbourhood_map is not None:
+        return neighbourhood_map[node_label]
+    elif network.rotation_system is not None:
+        return network.rotation_system[node_label]
+    else:
+        return [i for i in incoming_connections]
+
+
+def _get_memoized(activity_rule, ctx, memoization_table, memoization_key):
+    key = memoization_key.to_key(ctx)
+    if key in memoization_table:
+        return memoization_table[key]
+    else:
+        result = activity_rule(ctx)
+        result = check_np(result)
+        memoization_table[key] = result
+        return result
+
+
+class MemoizationKey:
+    def to_key(self, ctx):
+        pass
+
+
+class _OrderedMemoizationKey(MemoizationKey):
+    """
+    An implementation of a memoization key that simply converts the context's neighbourhood activities into
+    a tuple. This assumes that the ordering of the nodes in the neighbourhood matters, and that the ordering has
+    been specified, such as by setting a Rotation System on the network. For example, a CTRBL rule, or an ECA rule,
+    would require a Rotation System on the network.
+
+    Note that this implementation doesn't include the current node if it isn't in its neighbourhood. If the rule
+    depends on the current node, but the current node isn't in its neighbourhood, then the user should supply their
+    own memoization key that adds the current node's activity. If the rule doesn't depend on the current node's
+    activity, then we'd be adding cache entries unnecessarily if we added the current node to the key.
+    e.g. [1,0,2], [1,3,2], [1,5,2], etc. all map to 9, so all we really need is a cache entry with [1,2]->9, but
+    instead we'd end up adding the entries [0,1,2]->9, [3,1,2]->9, [5,1,2]->9, etc.
+    """
+    def to_key(self, ctx):
+        return tuple(ctx.neighbourhood_activities)
+
+
+class _UnorderedMemoizationKey(MemoizationKey):
+    """
+    An implementation of a memoization key that simply converts the context's neighbourhood activities into
+    a frozenset of item counts. This assumes that the ordering of the nodes in the neighbourhood doesn't matter.
+    For example, a totalistic 1D CA rule does not require an ordering of the neighbourhood to be specified.
+    It would be less efficient to just return a tuple of the activities here, since we'd end up with needless
+    entries in the cache. The rule doesn't depend on node order, so states like (1,0,1) and (0,1,1) should result
+    in the same cache hit, and evaluate to the same result.
+
+    Note that this implementation doesn't include the current node if it isn't in its neighbourhood. If the rule
+    depends on the current node, but the current node isn't in its neighbourhood, then the user should supply their
+    own memoization key that adds the current node's activity. If the rule doesn't depend on the current node's
+    activity, then we'd be adding cache entries unnecessarily if we added the current node to the key.
+    e.g. [1,0,2], [2,3,1], [2,5,1], etc. all map to 9, so all we really need is a cache entry with {1,2}->9, but
+    instead we'd end up adding the entries (0,{1,2})->9, (3,{1,2})->9, (5,{1,2})->9, etc.
+    """
+    def to_key(self, ctx):
+        return frozenset(collections.Counter(ctx.neighbourhood_activities).items())
+
+
 def evolve_topology(topology_rule, t, activities, prev_network, trajectory, copy_network,
                     persist_network, added_nodes=None, removed_nodes=None):
     network = prev_network
@@ -309,6 +398,12 @@ def _get_past_activities(past_conditions, trajectory, t):
     return None
 
 
+def _get_neighbourhood_map(network):
+    if network.rotation_system is not None:
+        return network.rotation_system
+    return {node_label: [i for i in network.in_edges(node_label)] for node_label in network.nodes}
+
+
 def _get_input_function(timesteps=None, input=None):
     if timesteps is None and input is None:
         raise Exception("either the timesteps or input must be provided")
