@@ -76,7 +76,8 @@ def plot_grid_multiple(ca_list, shape=None, slice=-1, titles=None, colormap='Gre
 
 
 def animate_activities(trajectory_or_activities, title='', shape=None, save=False, interval=50, colormap='Greys',
-                       vmin=None, vmax=None, show_grid=False, show_margin=True, scale=0.6, dpi=80):
+                       vmin=None, vmax=None, show_grid=False, show_margin=True, scale=0.6, dpi=80, blit=True,
+                       with_timestep=False):
     if len(trajectory_or_activities) is 0:
         raise Exception("there are no activities")
     if isinstance(trajectory_or_activities[0], State):
@@ -88,7 +89,7 @@ def animate_activities(trajectory_or_activities, title='', shape=None, save=Fals
         activities = _reshape_for_animation(activities, shape)
     cmap = plt.get_cmap(colormap)
     fig, ax = plt.subplots()
-    plt.title(title)
+    title_text = plt.title(title)
     if not show_margin:
         fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
 
@@ -117,8 +118,10 @@ def animate_activities(trajectory_or_activities, title='', shape=None, save=Fals
         if i['index'] == len(activities):
             i['index'] = 0
         im.set_array(activities[i['index']])
-        return im, grid
-    ani = animation.FuncAnimation(fig, updatefig, interval=interval, blit=True, save_count=len(activities))
+        if with_timestep:
+            title_text.set_text("timestep: %s" % (i['index']+1))
+        return im, grid, title_text
+    ani = animation.FuncAnimation(fig, updatefig, interval=interval, blit=blit, save_count=len(activities))
     if save:
         ani.save('evolved.gif', dpi=dpi, writer="imagemagick")
     plt.show()
@@ -170,12 +173,19 @@ def plot_network(network, layout="shell", with_labels=True, node_color="#1f78b4"
 
 
 def animate_network(trajectory, save=False, interval=50, dpi=80, layout="shell",
-                    with_labels=True, node_color="b", node_size=30):
+                    with_labels=True, with_arrows=True, node_color="b", node_size=30, with_timestep=False, show=True):
     fig, ax = plt.subplots()
 
-    def update(state):
+    def update(arg):
         ax.clear()
 
+        i, state = arg
+
+        if with_timestep:
+            ax.set_title("timestep: %s" % (i+1))
+
+        color = node_color[i] if type(node_color) == dict else node_color
+
         network = state.network
         G = nx.MultiDiGraph()
         for n in network.nodes:
@@ -183,20 +193,29 @@ def animate_network(trajectory, save=False, interval=50, dpi=80, layout="shell",
         for edge in network.edges:
             G.add_edge(edge[0], edge[1])
 
-        if layout == "shell":
-            nx.draw_shell(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
-        elif layout == "spring":
-            nx.draw_spring(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
-        elif isinstance(layout, dict):
-            nx.draw(G, pos=layout, with_labels=with_labels, node_color=node_color, node_size=node_size)
-        else:
-            raise Exception("unsupported layout: %s" % layout)
+        state_layout = layout[i] if type(layout) == list or type(layout) == tuple else layout
 
-    ani = animation.FuncAnimation(fig, update, frames=trajectory, interval=interval,
+        if state_layout == "shell":
+            nx.draw_shell(G, with_labels=with_labels, node_color=color, node_size=node_size, arrows=with_arrows)
+        elif state_layout == "spring":
+            nx.draw_spring(G, with_labels=with_labels, node_color=color, node_size=node_size, arrows=with_arrows)
+        elif state_layout == "planar":
+            nx.draw_planar(G, with_labels=with_labels, node_color=color, node_size=node_size, arrows=with_arrows)
+        elif state_layout == "kamada-kawai":
+            nx.draw_kamada_kawai(G, with_labels=with_labels, node_color=color, node_size=node_size, arrows=with_arrows)
+        elif isinstance(state_layout, dict):
+            nx.draw(G, pos=state_layout, with_labels=with_labels, node_color=color, node_size=node_size, arrows=with_arrows)
+        else:
+            raise Exception("unsupported layout: %s" % state_layout)
+
+    ani = animation.FuncAnimation(fig, update, frames=list(enumerate(trajectory)), interval=interval,
                                   save_count=len(trajectory))
     if save:
         ani.save('evolved.gif', dpi=dpi, writer="imagemagick")
-    plt.show()
+    if show:
+        plt.show()
+
+    return ani
 
 
 def plot1D(x, y, color=None, label=None, xlabel=None, ylabel=None, xlim=None, ylim=None, twinx=False,
